/*
 *	libxtadm
 *	Copyright © Jan Engelhardt, 2012
 *
 *	This program is free software: you can redistribute it and/or
 *	modify it under the terms of the GNU Lesser General Public License
 *	as published by the Free Software Foundation, either version 2.1 of
 *	the License, or (at your option) any later version.
 */
#include <errno.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libnetfilter_xtables/xtnetlink.h>
#include <libnetfilter_xtables/xtuser.h>
#include <xtables.h>
#include <netinet/in.h>
#include <linux/netfilter.h>
#include "internal.h"

enum {
	NF_MAX_HOOKS = 8,
};

static bool xtadm_promote_parse(struct xtadm_state *state,
                                struct xtnetlink_promote_param *promo)
{
	unsigned int nfproto;
	char *e;

	if (state->promote_nfproto == NULL && state->promote_hook == NULL &&
	    state->promote_priority == NULL)
		return true;
	if (!(state->promote_nfproto != NULL &&
	      state->promote_hook != NULL &&
	      state->promote_priority != NULL)) {
		fprintf(stderr, "ERROR: Either all or none of --nfproto, "
		        "--hook and --priority must be specified for -N.\n");
		return false;
	}
	if (!xtables_strtoui(state->promote_nfproto, NULL, &nfproto,
	    0, NFPROTO_NUMPROTO)) {
		fprintf(stderr, "ERROR: Could not understand "
		        "--nfproto \"%s\"\n", state->promote_nfproto);
		return false;
	}
	promo->nfproto = nfproto;
	if (!xtables_strtoui(state->promote_hook, NULL, &promo->hooknum,
	    0, NF_MAX_HOOKS)) {
		fprintf(stderr, "ERROR: Could not understand "
		        "--hook \"%s\"\n", state->promote_hook);
		return false;
	}
	promo->priority = strtol(state->promote_priority, &e, 0);
	if (e == state->promote_priority || *e != '\0') {
		fprintf(stderr, "ERROR: Could not understand "
		        "--priority \"%s\"\n", state->promote_priority);
		return false;
	}
	return true;
}

int xtadm_do_demote(struct xtadm_state *state)
{
	int ret;

	ret = xtnetlink_chain_demote(xtnl(state), state->chain_name);
	if (ret != 0)
		fprintf(stderr, "ERROR: Chain demotion failed: %s\n",
		        xtuser_strerror(ret));
	return (ret == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}

int xtadm_do_promote(struct xtadm_state *state)
{
	struct xtnetlink_promote_param promo = {.nfproto = NFPROTO_NUMPROTO};
	int ret;

	if (!xtadm_promote_parse(state, &promo))
		return EXIT_FAILURE;
	if (promo.nfproto == NFPROTO_NUMPROTO) {
		fprintf(stderr, "ERROR: Promote paramters need to be specified.\n");
		return EXIT_FAILURE;
	}

	promo.name = state->chain_name;
	ret = xtnetlink_chain_promote(xtnl(state), &promo);
	if (ret != 0)
		fprintf(stderr, "ERROR: Chain promotion failed: %s\n",
		        xtuser_strerror(ret));
	return (ret == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}

int xtadm_do_newchain(struct xtadm_state *state)
{
	struct xtnetlink_promote_param promo = {.nfproto = NFPROTO_NUMPROTO};
	struct xtuser_chain *chain;
	int ret;

	if (!xtadm_promote_parse(state, &promo))
		/*
		 * Make sure promo parameters are right before attempting
		 * any creations.
		 */
		return EXIT_FAILURE;

	chain = xtuser_chain_create(NULL, state->chain_name);
	if (chain == NULL) {
		fprintf(stderr, "ERROR: Chain object creation failed: %s\n",
		        xtuser_strerror(-errno));
		return EXIT_FAILURE;
	}
	ret = xtnetlink_chain_create(xtnl(state), chain);
	xtuser_chain_free(chain);
	if (ret != 0) {
		fprintf(stderr, "ERROR: Chain creation failed: %s\n",
		        xtuser_strerror(ret));
		return EXIT_FAILURE;
	}
	if (promo.nfproto == NFPROTO_NUMPROTO)
		return (ret == 0) ? EXIT_SUCCESS : EXIT_FAILURE;

	/* --nfproto was given */
	ret = xtadm_do_promote(state);
	if (ret != EXIT_SUCCESS)
		xtadm_do_delchain(state);
	return ret;
}

static int xtadm_do_delallchains(struct xtadm_state *state)
{
	struct xtuser_table *table;
	int ret;

	table = xtuser_table_new();
	if (table == NULL) {
		fprintf(stderr, "ERROR: Table object creation failed: %s\n",
		        strerror(errno));
		return EXIT_FAILURE;
	}
	ret = xtnetlink_table_replace(xtnl(state), table);
	if (ret != 0)
		fprintf(stderr, "ERROR: Chain genocide failed: %s\n",
		        xtuser_strerror(ret));
	xtuser_table_free(table);
	return (ret == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}

int xtadm_do_delchain(struct xtadm_state *state)
{
	int ret;

	if (state->chain_name == NULL)
		return xtadm_do_delallchains(state);

	ret = xtnetlink_chain_del(xtnl(state), state->chain_name);
	if (ret != 0) {
		fprintf(stderr, "ERROR: Chain deletion failed: %s\n",
		        xtuser_strerror(ret));
		return EXIT_FAILURE;
	}
	return EXIT_SUCCESS;
}

static int xtadm_do_flushall(struct xtadm_state *state)
{
	struct xtuser_table *table;
	struct xtuser_table_iter *iter;
	struct xtuser_chain *chain;
	int ret;

	table = xtnetlink_table_dump(xtnl(state));
	if (table == NULL) {
		fprintf(stderr, "ERROR: Table dump failed: %s\n",
		        strerror(errno));
		return EXIT_FAILURE;
	}
	iter = xtuser_table_travinit(table);
	if (iter == NULL) {
		ret = -errno;
		fprintf(stderr, "ERROR: %s\n", xtuser_strerror(ret));
		xtuser_table_free(table);
		return ret;
	}
	while ((chain = xtuser_table_traverse(iter)) != NULL)
		xtuser_chain_flush(chain);
	xtuser_table_travfree(iter);
	ret = xtnetlink_table_replace(xtnl(state), table);
	if (ret != 0)
		fprintf(stderr, "ERROR: Table replace failed: %s\n",
		        xtuser_strerror(ret));
	xtuser_table_free(table);
	return (ret == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}

int xtadm_do_flush(struct xtadm_state *state)
{
	struct xtuser_chain *chain;
	int ret;

	if (state->chain_name == NULL)
		return xtadm_do_flushall(state);

	chain = xtuser_chain_create(NULL, state->chain_name);
	if (chain == NULL) {
		fprintf(stderr, "ERROR: Chain object creation failed: %s\n",
		        strerror(errno));
		return EXIT_FAILURE;
	}
	ret = xtnetlink_chain_splice(xtnl(state), chain, 0, XTNL_CHAIN_FLUSH);
	if (ret != 0)
		fprintf(stderr, "ERROR: Chain flush failed: %s\n",
		        xtuser_strerror(ret));
	xtuser_chain_free(chain);
	return (ret == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}
