/*
 *	libxtadm
 *	Copyright © Jan Engelhardt, 2012
 *
 *	This program is free software: you can redistribute it and/or
 *	modify it under the terms of the GNU Lesser General Public License
 *	as published by the Free Software Foundation, either version 2.1 of
 *	the License, or (at your option) any later version.
 */
#include <errno.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libHX/init.h>
#include <libnetfilter_xtables/xtnetlink.h>
#include <libnetfilter_xtables/xtuser.h>
#include <libxtadm/xtadm.h>
#include <xtables.h>
#include "internal.h"

static pthread_mutex_t xtadm_init_lock = PTHREAD_MUTEX_INITIALIZER;
static unsigned long xtadm_use_count;

static void __attribute__((noreturn))
xtadm_exit_err(enum xtables_exittype status, const char *msg, ...)
{
	va_list argp;
	va_start(argp, msg);
	vfprintf(stderr, msg, argp);
	va_end(argp);
	fprintf(stderr, "\n# Old exit interface.\n");
	exit(1);
}

static struct xtables_globals xtadm_ugly_global_params = {
	.exit_err = xtadm_exit_err,
};

/**
 * Initialize the library. This is refcounted and properly nests, and all
 * libxtadm (including other shared libraries that make use of our functions)
 * must call this.
 */
EXPORT_SYMBOL int xtadm_init(void)
{
	int ret;

	pthread_mutex_lock(&xtadm_init_lock);
	if (xtadm_use_count == 0) {
		xtables_set_params(&xtadm_ugly_global_params);
		ret = HX_init();
		if (ret <= 0)
			goto out;
		ret = xtuser_init();
		if (ret < 0)
			goto out2;
	}
	xtables_init();
	xtables_set_nfproto(NFPROTO_UNSPEC);
#if defined(ALL_INCLUSIVE) || defined(NO_SHARED_LIBS)
	init_extensions();
	init_extensions4();
	init_extensions6();
#endif
	++xtadm_use_count;
	pthread_mutex_unlock(&xtadm_init_lock);
	return 1;
 out2:
	HX_exit();
 out:
	return ret;
}

/**
 * Drop the library reference count. Will deinitialize all the library-specific
 * internal data structures once possible.
 */
EXPORT_SYMBOL void xtadm_exit(void)
{
	pthread_mutex_lock(&xtadm_init_lock);
	if (xtadm_use_count == 0) {
		fprintf(stderr, "%s: reference count is already zero!\n",
		        __func__);
	} else if (xtadm_use_count == 1) {
		--xtadm_use_count;
		xtuser_exit();
		HX_exit();
	} else {
		--xtadm_use_count;
	}
	pthread_mutex_unlock(&xtadm_init_lock);
}

/**
 * @xtadmp:	place to store xtadm handle
 *
 * Opens the connection to the Netlink socket. Returns an "xtadm" handle
 * in @xtadmp, and 0 as value on success, otherwise negative error.
 */
EXPORT_SYMBOL int xtadm_open(struct xtadm_handle **xtadmp)
{
	struct xtadm_handle *adm;
	int ret;

	adm = *xtadmp = malloc(sizeof(*adm));
	if (adm == NULL)
		return -errno;

	adm->xtnl_handle = xtnetlink_open(XTNL_NO_FLAGS);
	if (adm->xtnl_handle == NULL) {
		ret = -errno;
		fprintf(stderr, "ERROR: xtnetlink_open: %s\n",
		        strerror(errno));
		free(adm);
		return ret;
	}

	return 0;
}

/**
 * @adm:	xtadm data to free
 *
 * Release an xtadm handle and free its resources.
 */
EXPORT_SYMBOL void xtadm_close(struct xtadm_handle *adm)
{
	xtnetlink_close(adm->xtnl_handle);
	free(adm);
}
