/*
 *	libxtadm
 *	Copyright © Jan Engelhardt, 2012
 *
 *	This program is free software: you can redistribute it and/or
 *	modify it under the terms of the GNU Lesser General Public License
 *	as published by the Free Software Foundation, either version 2.1 of
 *	the License, or (at your option) any later version.
 */
#include <errno.h>
#include <getopt.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <libHX/ctype_helper.h>
#include <libHX/defs.h>
#include <libHX/deque.h>
#include <libHX/misc.h>
#include <libHX/string.h>
#include <libnetfilter_xtables/libnetfilter_xtables.h>
#include <libxtadm/xtadm.h>
#include <xtables.h>
#include <linux/netfilter.h>
#include <linux/netfilter_arp.h>
#include <linux/netfilter_bridge.h>
#include "internal.h"

enum {
	XT_OPTION_OFFSET_SCALE = 256,
};

static void xtadm_state_free(struct xtadm_state *state)
{
	free(state->opts);
}

static void xtadm_state_reset(struct xtadm_state *state)
{
	struct xtadm_handle *ah = state->adm_handle;

	xtadm_state_free(state);
	memset(state, 0, sizeof(*state));
	state->action = XTADM_ACTION_NONE;
	state->adm_handle = ah;
}

static int xtadm_no_command_specified(void)
{
	fprintf(stderr, "libxtadm: no command specified\n");
	fprintf(stderr, "See xtadm(8) for more information.\n");
	/* The C interface uses the same options as xtadm(8). */
	return EXIT_FAILURE;
}

static int xtadm_finish_mt(struct xtadm_state *state)
{
	const struct xtables_match *m  = state->match_mod;
	const struct xtables_target *t = state->target_mod;
	int ret = 0;

	if (m != NULL) {
		xtables_option_mfcall(m);
		ret = xtuser_rule_push_match(state->rule, m->name,
		      m->revision, m->m->data, m->size);
	}
	if (t != NULL) {
		xtables_option_tfcall(t);
		ret = xtuser_rule_push_target(state->rule, t->name,
		      t->revision, t->t->data, t->size);
	}
	state->mt_name    = NULL;
	state->match_mod  = NULL;
	state->target_mod = NULL;
	if (ret != 0)
		fprintf(stderr, "ERROR: %s: %s\n", __func__,
		        xtuser_strerror(ret));
	return ret;
}

static int xtadm_do_command(struct xtadm_state *state)
{
	int ret;

	ret = xtadm_finish_mt(state);
	if (ret != 0)
		return ret;
	switch (state->action) {
	case XTADM_ACTION_ADD:
		ret = xtadm_do_add(state);
		break;
	case XTADM_ACTION_DELETE:
		ret = xtadm_do_delete(state);
		break;
	case XTADM_ACTION_FLUSH:
		ret = xtadm_do_flush(state);
		break;
	case XTADM_ACTION_INSERT:
		ret = xtadm_do_insert(state);
		break;
	case XTADM_ACTION_NEWCHAIN:
		ret = xtadm_do_newchain(state);
		break;
	case XTADM_ACTION_REPLACE:
		ret = xtadm_do_replace(state);
		break;
	case XTADM_ACTION_SHOW:
		ret = xtadm_do_show(state);
		break;
	case XTADM_ACTION_DELCHAIN:
		ret = xtadm_do_delchain(state);
		break;
	case XTADM_ACTION_PROMOTE:
		ret = xtadm_do_promote(state);
		break;
	case XTADM_ACTION_DEMOTE:
		ret = xtadm_do_demote(state);
		break;
	case XTADM_ACTION_EXECFILE:
		ret = xtadm_execf(state->adm_handle, state->exec_file);
		break;
	default:
		return xtadm_no_command_specified();
	}
	xtadm_state_reset(state);
	return ret;
}

static int xtadm_set_rule_num(struct xtadm_state *state, const char *arg)
{
	switch (state->action) {
	case XTADM_ACTION_DELETE:
	case XTADM_ACTION_INSERT:
	case XTADM_ACTION_REPLACE:
	case XTADM_ACTION_SHOW:
		break;
	default:
		fprintf(stderr, "ERROR: Command -[DIRS] must precede -#");
		return EXIT_FAILURE;
	}
	if (!xtables_strtoui(arg, NULL, &state->rule_num, 0, UINT_MAX)) {
		fprintf(stderr, "ERROR: \"%s\" not recognized as a number\n", arg);
		return EXIT_FAILURE;
	}
	return 0;
}

static bool xtadm_in_mt_action(struct xtadm_state *state)
{
	switch (state->action) {
	case XTADM_ACTION_ADD:
	case XTADM_ACTION_CHECK:
	case XTADM_ACTION_DELETE:
	case XTADM_ACTION_INSERT:
	case XTADM_ACTION_REPLACE:
		return true;
	default:
		return false;
	}
}

static int xtadm_start_cmd(struct xtadm_state *state,
                           enum xtadm_action action, const char *arg)
{
	int ret = 0;

	if (state->action != XTADM_ACTION_NONE) {
		ret = xtadm_do_command(state);
		if (ret < 0)
			return ret;
		/* implicit clean by do_command */
	}

	state->action = action;
	if (action == XTADM_ACTION_EXECFILE)
		state->exec_file = arg;
	else
		state->chain_name = arg;
	if (xtadm_in_mt_action(state)) {
		state->rule = xtuser_rule_new();
		if (state->rule == NULL) {
			fprintf(stderr, "ERROR: xtuser_rule_new: %s\n",
			        strerror(errno));
			return -errno;
		}
	}
	return ret;
}

static int xtadm_match_init(struct xtables_match *mod)
{
	int ret;

	if (mod->udata_size != 0) {
		/*
		 * As soon as a subsequent instance of the same match
		 * is used, e.g. "-m time -m time", the first instance
		 * is no longer reachable anyway, so we can free udata.
		 * Same goes for target.
		 */
		free(mod->udata);
		mod->udata = calloc(1, mod->udata_size);
		if (mod->udata == NULL) {
			ret = -errno;
			fprintf(stderr, "ERROR: malloc: %s\n", strerror(errno));
			return ret;
		}
	}
	if (mod->init != NULL)
		mod->init(mod->m);
	return 0;
}

static int xtadm_target_init(struct xtables_target *mod)
{
	int ret;

	if (mod->udata_size != 0) {
		free(mod->udata);
		mod->udata = calloc(1, mod->udata_size);
		if (mod->udata == NULL) {
			ret = -errno;
			fprintf(stderr, "ERROR: malloc: %s\n", strerror(errno));
			return ret;
		}
	}
	if (mod->init != NULL)
		mod->init(mod->t);
	return 0;
}

static int xtadm_start_match(struct xtadm_state *state, const char *arg)
{
	struct xtables_match *mod;
	struct option *newopts = NULL;
	int ret;

	if (!xtadm_in_mt_action(state)) {
		fprintf(stderr, "ERROR: -m can only be used from -ACDIR commands\n");
		return EXIT_FAILURE;
	}
	ret = xtadm_finish_mt(state);
	if (ret != 0)
		return ret;

	state->mt_name = arg;
	mod = xtables_find_match(arg, XTF_TRY_LOAD, NULL);
	if (mod == NULL) {
		fprintf(stderr, "ERROR: Plugin for match \"%s\" not found\n", arg);
		return -ENOENT;
	}

	state->match_mod  = mod;
	state->target_mod = NULL;
	mod->m = malloc(sizeof(struct xt_entry_match) + mod->size);
	if (mod->m == NULL) {
		perror("malloc");
		return -errno;
	}
	if (mod->real_name != NULL) {
		state->mt_name = mod->real_name;
		fprintf(stderr, "WARNING: The \"%s\" match is obsolete. "
		        "Use \"%s\" instead.\n", mod->name, mod->real_name);
	}
	ret = xtadm_match_init(mod);
	if (ret != 0)
		return EXIT_FAILURE;
	if (mod->x6_options != NULL)
		newopts = xtables_options_xfrm(NULL, state->opts,
		          mod->x6_options, &mod->option_offset);
	else if (mod->extra_opts != NULL)
		newopts = xtables_merge_options(NULL, state->opts,
		          mod->extra_opts, &mod->option_offset);
	if (newopts != state->opts) {
		free(state->opts);
		state->opts = newopts;
	}
	return 0;
}

static int xtadm_start_target(struct xtadm_state *state, const char *arg)
{
	struct xtables_target *mod;
	struct option *newopts = NULL;
	int ret;

	if (!xtadm_in_mt_action(state)) {
		fprintf(stderr, "ERROR: -m can only be used from -ACDIR commands\n");
		return EXIT_FAILURE;
	}
	state->mt_name = arg;
	mod = xtables_find_target(arg, XTF_TRY_LOAD);
	if (mod == NULL) {
		fprintf(stderr, "ERROR: Plugin for target \"%s\" not found\n", arg);
		return -ENOENT;
	}

	state->match_mod  = NULL;
	state->target_mod = mod;
	mod->t = malloc(sizeof(struct xt_entry_target) + mod->size);
	if (mod->t == NULL) {
		perror("malloc");
		return -errno;
	}
	if (mod->real_name != NULL) {
		state->mt_name = mod->real_name;
		fprintf(stderr, "WARNING: The \"%s\" target is obsolete. "
		        "Use \"%s\" instead.\n", mod->name, mod->real_name);
	}
	ret = xtadm_target_init(mod);
	if (ret != 0)
		return EXIT_FAILURE;
	if (mod->x6_options != NULL)
		newopts = xtables_options_xfrm(NULL,
		          state->opts, mod->x6_options, &mod->option_offset);
	else if (mod->extra_opts != NULL)
		newopts = xtables_merge_options(NULL,
		          state->opts, mod->extra_opts, &mod->option_offset);
	if (newopts != state->opts) {
		free(state->opts);
		state->opts = newopts;
	}
	return 0;
}

static int xtadm_set_invert(struct xtadm_state *state)
{
	if (state->invert) {
		fprintf(stderr, "ERROR: Ain't nobody doing no double negations!\n");
		return EXIT_FAILURE;
	}
	state->invert = true;
	return EXIT_SUCCESS;
}

static int
xtadm_set_promote(struct xtadm_state *state, char op, const char *arg)
{
	if (state->action != XTADM_ACTION_NEWCHAIN &&
	    state->action != XTADM_ACTION_PROMOTE) {
		fprintf(stderr, "ERROR: --hook can only be used with the "
		        "-N or --promote commands.\n");
		return EXIT_FAILURE;
	}
	/* Just copy the pointer; it won't change until the command is done. */
	if (op == 'h')
		state->promote_hook = arg;
	else if (op == 'f')
		state->promote_nfproto = arg;
	else if (op == 'p')
		state->promote_priority = arg;
	return 0;
}

static int
xtadm_dynamic_option(struct xtadm_state *state, int c, const char *arg)
{
	struct xtables_match *m  = state->match_mod;
	struct xtables_target *t = state->target_mod;
	const char *argarray[2] = {arg, NULL};

	if (t != NULL && (t->parse != NULL || t->x6_parse != NULL) &&
	    c >= t->option_offset &&
	    c < t->option_offset + XT_OPTION_OFFSET_SCALE) {
		xtables_option_tpcall(c, const_cast2(char **, argarray),
		                      state->invert, t, NULL);
		return 0;
	}
	if (m != NULL && (m->parse != NULL || m->x6_parse != NULL) &&
	    c >= m->option_offset &&
	    c < m->option_offset + XT_OPTION_OFFSET_SCALE) {
		xtables_option_mpcall(c, const_cast2(char **, argarray),
		                      state->invert, m, NULL);
		return 0;
	}
	return '?';
}

enum {
	CL_OPT_NFPROTO = 0x80,
	CL_OPT_HOOK,
	CL_OPT_PRIORITY,
	CL_CMD_PROMOTE,
	CL_CMD_DEMOTE,
	CL_CMD_HELP,
};

static const char xtadm_option_string[] =
	"-#:A:B::D:E:F::I:L::N:R:S::X::hm:j:";
static const struct option xtadm_long_options[] = {
	{.name = "nfproto",  .has_arg = true, .val = CL_OPT_NFPROTO},
	{.name = "hook",     .has_arg = true, .val = CL_OPT_HOOK},
	{.name = "priority", .has_arg = true, .val = CL_OPT_PRIORITY},
	{.name = "promote",  .has_arg = true, .val = CL_CMD_PROMOTE},
	{.name = "demote",   .has_arg = true, .val = CL_CMD_DEMOTE},
	{.name = "help",     .has_arg = true, .val = 'h'},
	{.name = NULL},
};

static inline bool xtadm_s_negation(const char *s)
{
	return s[0] == '!' && s[1] == '\0';
}

static inline bool xtadm_s_option(const char *s)
{
	if (s[0] != '-')
		return false; /* non option */
	if (s[1] == '\0')
		return false; /* lone dash */
	if (s[1] == '-' && s[2] == '\0')
		return false; /* option stopper */
	return true;
}

/**
 * @adm:	xtadm handle
 * @argc:	length of @argv
 * @argv:	vector of arguments, must be NULL-terminated
 *
 * Executes the text-based xtadm commands. Use this pretty much like you did
 * iptables(8). @argv[0] is a dummy field and ignored; options are to start
 * at @argv[1].
 */
EXPORT_SYMBOL int
xtadm_execvn(struct xtadm_handle *adm, int argc, char *const *argv)
{
	struct xtadm_state state;
	int c, ret = 0, saved_optind;

	if (argc < 0)
		argc = HX_zveclen(static_cast(const void *, argv));

	memset(&state, 0, sizeof(state));
	state.adm_handle = adm;
	state.opts = HX_memdup(xtadm_long_options, sizeof(xtadm_long_options));
	if (state.opts == NULL)
		return -errno;

	saved_optind = optind; /* globals suck */
	optind = 0;

	while ((c = getopt_long(argc, argv, xtadm_option_string,
	       state.opts, NULL)) >= 0)
	{
		if (optarg == NULL && argv[optind] != NULL &&
		    !xtadm_s_negation(argv[optind]) &&
		    !xtadm_s_option(argv[optind]))
			/* Collect optional argument */
			optarg = argv[optind++];

		switch (c) {
		case '#':
			ret = xtadm_set_rule_num(&state, optarg);
			break;
		case 'A':
			ret = xtadm_start_cmd(&state, XTADM_ACTION_ADD, optarg);
			break;
		case 'B':
			ret = xtadm_start_cmd(&state, XTADM_ACTION_EXECFILE, optarg);
			break;
		case 'C':
			ret = xtadm_start_cmd(&state, XTADM_ACTION_CHECK, optarg);
			break;
		case 'D':
			ret = xtadm_start_cmd(&state, XTADM_ACTION_DELETE, optarg);
			break;
		case 'F':
			ret = xtadm_start_cmd(&state, XTADM_ACTION_FLUSH, optarg);
			break;
		case 'I':
			ret = xtadm_start_cmd(&state, XTADM_ACTION_INSERT, optarg);
			break;
		case 'N':
			ret = xtadm_start_cmd(&state, XTADM_ACTION_NEWCHAIN, optarg);
			break;
		case 'R':
			ret = xtadm_start_cmd(&state, XTADM_ACTION_REPLACE, optarg);
			break;
		case 'L':
		case 'S':
			ret = xtadm_start_cmd(&state, XTADM_ACTION_SHOW, optarg);
			break;
		case 'X':
			ret = xtadm_start_cmd(&state, XTADM_ACTION_DELCHAIN, optarg);
			break;
		case 'm':
			ret = xtadm_start_match(&state, optarg);
			break;
		case 'j':
			ret = xtadm_start_target(&state, optarg);
			break;
		case CL_CMD_PROMOTE:
			ret = xtadm_start_cmd(&state, XTADM_ACTION_PROMOTE, optarg);
			break;
		case CL_CMD_DEMOTE:
			ret = xtadm_start_cmd(&state, XTADM_ACTION_DEMOTE, optarg);
			break;
		case 1: /* non-option */
			if (optarg[0] == '!' && optarg[1] == '\0') {
				ret = xtadm_set_invert(&state);
				if (ret != EXIT_SUCCESS)
					return ret;
			}
			fprintf(stderr, "non-option seen: %s\n", optarg);
			break;
		case CL_OPT_NFPROTO:
			ret = xtadm_set_promote(&state, 'f', optarg);
			break;
		case CL_OPT_HOOK:
			ret = xtadm_set_promote(&state, 'h', optarg);
			break;
		case CL_OPT_PRIORITY:
			ret = xtadm_set_promote(&state, 'p', optarg);
			break;
		default:
			ret = xtadm_dynamic_option(&state, c, optarg);
			if (ret != '?')
				break;
			/* fallthrough */
		case '?':
		case 'h':
			ret = XTADM_EXIT_UNKNOWN_OPTION;
			goto out;
		}
		if (ret != 0)
			goto out;
	}

	xtadm_do_command(&state);
out:
	xtadm_state_free(&state);
	optind = saved_optind;
	return ret;
}

/**
 * @adm:	xtadm instance
 * @arg ...:	commands/options to xtadm which say what to do
 *
 * Executes the text-based xtadm commands known from iptables.
 *
 * Hm, we are still lacking a way to get the output (e.g. from -S)
 * by means other than capturing process output...
 */
EXPORT_SYMBOL int xtadm_execl(struct xtadm_handle *adm, ...)
{
	const char *arg;
	struct HXdeque *dq;
	void **argv;
	va_list argp;
	unsigned int argc;
	int ret;

	dq = HXdeque_init();
	if (dq == NULL)
		return -errno;
	if (HXdeque_push(dq, "<internal>") == NULL)
		goto out;

	va_start(argp, adm);
	while ((arg = va_arg(argp, const char *)) != NULL)
		if (HXdeque_push(dq, arg) == NULL)
			goto out;
	BUILD_BUG_ON(sizeof(char *) != sizeof(void *));
	argv = HXdeque_to_vec(dq, &argc);
	if (argv == NULL)
		goto out;
	ret = xtadm_execvn(adm, argc, reinterpret_cast(char *const *, argv));
	free(argv);
	HXdeque_genocide(dq);
	return ret;

 out:
	ret = -errno;
	if (dq->first != NULL)
		dq->first->ptr = NULL; /* the immutable <internal> */
	HXdeque_genocide(dq);
	return ret;
}

/**
 * @adm:	xtadm instance
 * @cmd:	single-line command with sh quoting syntax
 *
 * Single-string multi-argument invocation that will split the command
 * before proceeding on.
 */
EXPORT_SYMBOL int xtadm_exec(struct xtadm_handle *adm, const char *cmd)
{
	int argc, ret;
	char **argv;

	argc = xtadm_args_split(cmd, &argv);
	if (argc < 0)
		return argc;
	ret = xtadm_execvn(adm, argc, argv);
	xtadm_args_free(argv);
	return ret;
}

EXPORT_SYMBOL int xtadm_execf(struct xtadm_handle *adm, const char *file)
{
	hxmc_t *ln = NULL;
	FILE *fp;
	int ret = 0;

	if (strcmp(file, "-") == 0)
		fp = stdin;
	else
		fp = fopen(file, "r");
	if (fp == NULL) {
		fprintf(stderr, "ERROR: %s: Could not open %s: %s\n",
		        __func__, file, strerror(errno));
		return EXIT_FAILURE;
	}
	while (HX_getl(&ln, fp) != NULL) {
		if (*ln == '#')
			continue;
		HX_chomp(ln);
		HX_strrtrim(ln);
		HX_strltrim(ln);
		if (*ln == '\0')
			continue;
		if (HXmc_strpcat(&ln, "<internal> ") == NULL) {
			fprintf(stderr, "ERROR: %s\n", strerror(errno));
			ret = EXIT_FAILURE;
			break;
		}
		ret = xtadm_exec(adm, ln);
		if (ret != 0)
			break;
	}
	HXmc_free(ln);
	fclose(fp);
	return ret;
}

/**
 * Breaks down a restore-like single line into its arguments.
 * This is used for command lines received from the user which go to xtadm,
 * and is also meant to be used for users to split lines received from xtadm.
 */
EXPORT_SYMBOL int xtadm_args_split(const char *cmd, char ***argp)
{
	char **argv, *buf, *wcur, quote = '\0';
	int i, argc = 1;

	while (HX_isspace(*cmd))
		++cmd;
	buf = HX_strdup(cmd);
	if (buf == NULL)
		return -errno;

	/*
	 * Unquote and tokenize string according to some shell-style
	 * conventions. There are only two allocations: buf and the argv array;
	 * this is the same approach as for a program's arguments as seen in
	 * a main() function.
	 */
	for (wcur = buf; *cmd != '\0'; ) {
		if (quote != '\0') {
			if (cmd[0] == '\\' && cmd[1] != '\0')
				*wcur++ = *++cmd;
			else if (*cmd == quote)
				quote = '\0';
			else
				*wcur++ = *cmd;
		} else {
			if (HX_isspace(*cmd)) {
				*wcur++ = '\0';
				++argc;
				while (HX_isspace(*cmd))
					++cmd;
				continue;
			}
			*wcur++ = *cmd;
		}
		++cmd;
	}

	argv = malloc(sizeof(*argv) * (argc + 1));
	argv[0] = buf;
	for (i = 1; i < argc; ++i)
		argv[i] = argv[i-1] + strlen(argv[i-1]) + 1;
	argv[i] = NULL;
	*argp = argv;
	return argc;
}

EXPORT_SYMBOL void xtadm_args_free(char **argv)
{
	free(*argv);
	free(argv);
}
