/*
 *	libxtadm
 *	Copyright © Jan Engelhardt, 2012
 *
 *	This program is free software: you can redistribute it and/or
 *	modify it under the terms of the GNU Lesser General Public License
 *	as published by the Free Software Foundation, either version 2.1 of
 *	the License, or (at your option) any later version.
 */
#include <errno.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <libnetfilter_xtables/xtnetlink.h>
#include <libnetfilter_xtables/xtuser.h>
#include <xtables.h>
#include <netinet/in.h>
#include <linux/netfilter.h>
#include "internal.h"

enum {
	NF_MAX_HOOKS = 8,
};

static struct xtuser_chain *xtadm_build_chain(struct xtadm_state *state)
{
	struct xtuser_chain *chain;
	int ret;

	chain = xtuser_chain_create(NULL, state->chain_name);
	if (chain == NULL) {
		ret = -errno;
		fprintf(stderr, "ERROR: Chain object creation failed: %s\n",
		        xtuser_strerror(ret));
		errno = -ret;
		return NULL;
	}
	ret = xtuser_chain_push(chain, state->rule);
	if (ret != 0) {
		fprintf(stderr, "ERROR: %s\n", xtuser_strerror(ret));
		xtuser_chain_free(chain);
		errno = -ret;
		return NULL;
	}
	state->rule = NULL;
	return chain;
}

int xtadm_do_add(struct xtadm_state *state)
{
	state->rule_num = XTNL_CHAIN_APPEND;
	return xtadm_do_insert(state);
}

int xtadm_do_insert(struct xtadm_state *state)
{
	struct xtuser_chain *chain;
	int ret;

	chain = xtadm_build_chain(state);
	if (chain == NULL)
		return -errno;
	if (state->rule_num != 0 && state->rule_num != XTNL_CHAIN_APPEND)
		/*
		 * The command line interfaces uses 1-based indices.
		 * So, if it's non-zero, someone has used the -# option,
		 * in which case we need to adjust it for chain_splice.
		 */
		--state->rule_num;
	ret = xtnetlink_chain_splice(xtnl(state), chain, state->rule_num, 0);
	if (ret != 0)
		fprintf(stderr, "ERROR: Chain splice on %s failed: %s\n",
		        state->chain_name, xtuser_strerror(ret));
	xtuser_chain_free(chain);
	return ret;
}

int xtadm_do_delete(struct xtadm_state *state)
{
	struct xtuser_chain *chain;
	int ret;

	if (state->rule_num == 0) {
		fprintf(stderr, "ERROR: A rule number must be specified "
		        "with -#. Rule indices are counted from 1.\n");
		return EXIT_FAILURE;
	} else {
		--state->rule_num;
	}
	chain = xtuser_chain_create(NULL, state->chain_name);
	if (chain == NULL) {
		ret = -errno;
		fprintf(stderr, "ERROR: Chain object creation failed: %s\n",
		        xtuser_strerror(ret));
		return -errno;
	}
	ret = xtnetlink_chain_splice(xtnl(state), chain, state->rule_num, 1);
	if (ret != 0)
		fprintf(stderr, "ERROR: Chain splice on %s failed: %s\n",
		        state->chain_name, xtuser_strerror(ret));
	xtuser_chain_free(chain);
	return ret;
}

int xtadm_do_replace(struct xtadm_state *state)
{
	if (state->rule_num == 0)
		/* !! */;
	return 1;
}
