/*
 *	libxtadm
 *	Copyright © Jan Engelhardt, 2012
 *
 *	This program is free software: you can redistribute it and/or
 *	modify it under the terms of the GNU Lesser General Public License
 *	as published by the Free Software Foundation, either version 2.1 of
 *	the License, or (at your option) any later version.
 */
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <libnetfilter_xtables/xtnetlink.h>
#include <libnetfilter_xtables/xtuser.h>
#include <libnetfilter_xtables/props.h>
#include <xtables.h>
#include <netinet/in.h>
#include <linux/netfilter.h>
#include <linux/netfilter/x_tables.h>
#include "internal.h"

static void xtadm_print_verdict(const struct xtuser_action *action)
{
	unsigned int v = xtuser_actionprop_verdict(action);

	if (v == NF_ACCEPT)
		printf(" -j ACCEPT");
	else if (v == NF_STOP)
		printf(" <NF_STOP>");
	else if ((v & NF_VERDICT_MASK) == NF_DROP)
		printf(" -j DROP");
	else if ((v & NF_VERDICT_MASK) == NF_QUEUE)
		printf(" -j QUEUE --queue %u", v >> NF_VERDICT_QBITS);
	else
		printf(" <verdict(0x%x)>", v);
}

static int xtadm_print_match(const struct xtuser_action *action)
{
	const struct xtables_match *match;
	size_t dsize = xtuser_actionprop_dsize(action);
	const char *name = xtuser_actionprop_name(action);

	//printf(" -m %s.%u", name, xtuser_actionprop_revision(action));
	printf(" -m %s", name);
	match = xtables_find_match(name, XTF_TRY_LOAD, NULL);
	if (match == NULL) {
		fprintf(stderr, "\nCannot find plugin for match \"%s\": %s\n",
		        name, strerror(errno));
		printf(" <%zu bytes of data>", dsize);
		return -1;
	}
	if (match->save != NULL)
		match->save(NULL, xtadm_fake_xtentry(xtuser_actionprop_data(action)));
	return 0;
}

static int xtadm_print_target(const struct xtuser_action *action)
{
	const struct xtables_target *target;
	size_t dsize = xtuser_actionprop_dsize(action);
	const char *name = xtuser_actionprop_name(action);

	//printf(" -j %s.%u", name, xtuser_actionprop_revision(action));
	printf(" -j %s", name);
	target = xtables_find_target(name, XTF_TRY_LOAD);
	if (target == NULL) {
		fprintf(stderr, "\nCannot find plugin for target \"%s\": %s\n",
		        name, strerror(errno));
		printf(" <%zu bytes of data>", dsize);
		return -1;
	}
	if (target->save != NULL)
		target->save(NULL, xtadm_fake_xtentry(xtuser_actionprop_data(action)));
	return 0;
}

static void xtadm_print_jg(const struct xtuser_action *action, char prefix)
{
	printf(" -%c %s", prefix, xtuser_actionprop_name(action));
}

static int xtadm_print_rule(const struct xtuser_rule *rule)
{
	struct xtuser_rule_iter *iter;
	const struct xtuser_action *action;
	enum xtuser_action_type atype;

	iter = xtuser_rule_travinit(rule);
	while ((action = xtuser_rule_traverse(iter)) != NULL) {
		atype = xtuser_actionprop_type(action);
		if (atype == XTUSER_ACTION_VERDICT)
			xtadm_print_verdict(action);
		else if (atype == XTUSER_ACTION_MATCH)
			xtadm_print_match(action);
		else if (atype == XTUSER_ACTION_TARGET)
			xtadm_print_target(action);
		else if (atype == XTUSER_ACTION_JUMP)
			xtadm_print_jg(action, 'j');
		else if (atype == XTUSER_ACTION_GOTO)
			xtadm_print_jg(action, 'g');
		else
			printf(" <action(%u)>\n", atype);
	}
	xtuser_rule_travfree(iter);
	return 0;
}

static void xtadm_print_chain_head(const struct xtuser_chain *chain)
{
	unsigned int tmp;

	printf("-N %s", xtuser_chainprop_name(chain));
	tmp = xtuser_chainprop_hooknfproto(chain);
	if (tmp != XTUSER_INVALID_NFPROTO)
		printf(" --nfproto %u --hook %u --priority %d",
		       tmp, xtuser_chainprop_hooknum(chain),
		       xtuser_chainprop_hookprio(chain));
	printf("\n");
}

static int xtadm_print_chain(const struct xtuser_chain *chain)
{
	struct xtuser_chain_iter *iter;
	const struct xtuser_rule *rule;
	unsigned int rule_idx = 0;

	iter = xtuser_chain_travinit(chain);
	while ((rule = xtuser_chain_traverse(iter)) != NULL) {
		printf("-A %s -# %u", xtuser_chainprop_name(chain),
		       ++rule_idx);
		xtadm_print_rule(rule);
		printf("\n");
	}

	xtuser_chain_travfree(iter);
	return 0;
}

static int xtadm_show_table(struct xtadm_state *state)
{
	struct xtuser_table_iter *iter;
	struct xtuser_table *table;
	struct xtuser_chain *chain;
	int ret = 0;

	table = xtnetlink_table_dump(xtnl(state));
	if (table == NULL) {
		ret = -errno;
		fprintf(stderr, "xtnetlink_table_dump: %s\n",
		        xtuser_strerror(ret));
		return ret;
	}

	printf("-X\n");
	iter = xtuser_table_travinit(table);
	while ((chain = xtuser_table_traverse(iter)) != NULL)
		xtadm_print_chain_head(chain);
	xtuser_table_travfree(iter);

	iter = xtuser_table_travinit(table);
	while ((chain = xtuser_table_traverse(iter)) != NULL) {
		ret = xtadm_print_chain(chain);
		if (ret != 0)
			break;
	}
	xtuser_table_travfree(iter);

	xtuser_table_free(table);
	return ret;
}

static int xtadm_show_chain(struct xtadm_state *state)
{
	struct xtuser_chain *chain;
	int ret = 0;

	chain = xtnetlink_chain_dump(xtnl(state), state->chain_name);
	if (chain == NULL) {
		ret = -errno;
		fprintf(stderr, "xtnetlink_chain_dump: %s\n",
		        xtuser_strerror(ret));
		return ret;
	}

	printf("-X %s\n", xtuser_chainprop_name(chain));
	xtadm_print_chain_head(chain);
	ret = xtadm_print_chain(chain);
	xtuser_chain_free(chain);
	return ret;
}

int xtadm_do_show(struct xtadm_state *state)
{
	if (state->chain_name == NULL)
		return xtadm_show_table(state);
	else	
		return xtadm_show_chain(state);
}
